# $Id: fragroute.py,v 1.7 2005/03/14 05:47:26 dugsong Exp $

"""intercept, modify, and rewrite egress traffic destined for a specific host.
"""

__author__ = 'Dug Song <dugsong@monkey.org>'
__copyright__ = 'Copyright (c) 2004 Dug Song'
__license__ = 'GPL'
__url__ = 'http://monkey.org/~dugsong/fragroute/'
__version__ = '2.0'

import copy, optparse, os, signal, struct, sys, time, traceback, types
import dnet, dpkt, event

class _DstIf(object):
    def __init__(self, dst):
        # Lookup outbound interface configuration.
        ifent = dnet.intf().get_dst(dst)
        self.addr = ifent['addr']
        self.mtu = ifent['mtu']
        self.name = ifent['name']
        
        # Lookup destination MAC address.
        dmac = None
        for i in range(3):
            if dst not in self.addr:
                gw = dnet.route().get(dst)
                gw.bits = dnet.IP_ADDR_BITS
                dmac = dnet.arp().get(gw)
            else:
                dmac = dnet.arp().get(dst)
            if dmac:
                break
            # XXX - send throwaway probe packet to force ARP lookup
            ip = dpkt.ip.IP(src=dst.ip, dst=dst.ip, ttl=0)
            dnet.ip().send(str(ip))
            time.sleep(1)
        if not dmac:
            raise OSError, 'no route to host %s' % dst

        # XXX - delete any local ARP entry
        if dst in self.addr:
            dnet.arp().delete(dst)
        
        # Cache outbound Ethernet header.
        eth = dpkt.ethernet.Ethernet(src=ifent['link_addr'].eth, dst=dmac.eth)
        self.eth_hdr = str(eth)
        
        # Open outbound device.
        self.eth = dnet.eth(self.name)
    
    def send_ip(self, ip):
        self.eth.send(self.eth_hdr + str(ip))

class Fragroute(object):
    """Egress-intercepting userland packet proxy.
    """
    usage = '%prog [OPTIONS] <host>'
    
    def __init__(self, *args, **kwargs):
        """Override to add options to the self.op OptionParser."""
        self.rand = dnet.rand()
        self.op = optparse.OptionParser(usage=self.usage)
    
    def genfrags(self, buf, size, firstlen=0, overlap=None):
        """Generator to split a buffer into chunks of a certain size with an
        optional first chunk size, yielding an (offset, chunk) tuple.
        """
        if firstlen:
            yield 0, buf[:firstlen]
        offsets = range(firstlen, len(buf), size)
        pad = size / 2 or size
        for off in offsets:
            if overlap == 'left' and off > offsets[0]:
                yield off - pad, self.rand.get(pad) + buf[off:off+size]
            elif overlap == 'right' and off < offsets[-1] or \
                     (len(buf) - firstlen) % size > pad:
                yield off, buf[off:off+size] + self.rand.get(pad)
            else:
                yield off, buf[off:off+size]
            
    def ip_frag(self, ip, size=8, overlap=None):
        """Fragment IP packet (preserving transport protocol header in the
        first fragment to bypass filters that block "short" fragments),
        returning list of fragmented IP packets.
        """
        if ip.p == dpkt.ip.IP_PROTO_TCP:
            firstlen = max(ip.tcp.off << 2, size)
        else:
            firstlen = max(8, size)
        
        if firstlen & 0x7:
            firstlen = (firstlen & ~0x7) + 8
        
        frags = []
        if len(ip.data) > firstlen:
            for off, data in self.genfrags(str(ip.data), size, firstlen,
                                           overlap=overlap):
                ip2 = copy.copy(ip)
                ip2.off = dpkt.ip.IP_MF | (off >> 3)
                ip2.data = data
                ip2.len = 20 + len(data)
                ip2.sum = 0
                frags.append(ip2)
            frags[-1].off &= ~dpkt.ip.IP_MF
        else:
            frags.append(ip)

        return frags

    def tcp_seg(self, ip, size=4, overlap=None):
        """Segment TCP/IP packet, returning list of segments."""
        segs = []
        if ip.p == dpkt.ip.IP_PROTO_TCP and len(ip.tcp.data) > size:
            for off, data in self.genfrags(str(ip.tcp.data), size,
                                           overlap=overlap):
                ip2 = copy.deepcopy(ip)
                ip2.id = self.rand.uint16()
                ip2.len = ip.__hdr_len__ + (ip.tcp.off << 2) + len(data)
                ip2.tcp.seq += off
                ip2.tcp.data = data
                ip2.sum = ip2.tcp.sum = 0
                segs.append(ip2)
        else:
            segs.append(ip)
        return segs

    def transform(self, ip):
        """Transform IP packet, returning list of IP packets to resend.
        """
        raise NotImplementedError
    
    def handle_read(self, fd):
        try:
            buf = os.read(fd, self.dev.mtu)[4:]
            for ip in self.transform(dpkt.ip.IP(buf)):
                self.dev.send_ip(str(ip))
        except:
            traceback.print_exc()
        return True
    
    def handle_sig(self, sig):
        print >>sys.stderr, 'exiting on signal %d' % sig
        event.abort()
        
    def run(self, dst):
        # Save outbound device.
        self.dev = _DstIf(dst)
        
        # Bring up tunnel to dst.
        src = self.dev.addr
        src.bits = dnet.IP_ADDR_BITS
        self.tun = dnet.tun(src, dst, self.dev.mtu)
        
        print >>sys.stderr, 'listening on %s, sending on %s: dst %s' % \
              (self.tun.name, self.dev.name, dst)
        
        event.init()
        event.read(self.tun.fd, self.handle_read, self.tun.fd).add()
        event.signal(signal.SIGINT, self.handle_sig, signal.SIGINT).add()
        event.dispatch()
        del self.tun, self.dev
        
    def main(self, argv=None):
        # XXX - try Psyco
        try:
            import psyco
            psyco.full()
        except ImportError:
            pass

        self.opts, args = self.op.parse_args(argv or sys.argv[1:])
        if not args:
            self.op.error('missing host')
        elif len(args) > 1:
            self.op.error('only one host may be specified')
        
        self.run(dnet.addr(args[0]))
    
def main(argv=None):
    """Load and run any Fragroute subclass found in __main__ scope."""
    module = __import__('__main__')
    for name in dir(module):
        cls = getattr(module, name)
        if isinstance(cls, (type, types.ClassType)) and \
           issubclass(cls, Fragroute) and cls != Fragroute:
            o = cls()
            return o.main(argv)

