#!/usr/bin/env python

"""reassembly policy tests from U. Shankar and V. Paxson, "Active Mapping: Resisting NIDS Evasion Without Altering Traffic", Proc. IEEE Symposium on Security and Privacy, May 2003.
"""

import socket, sys
import dpkt, dnet, pcap

def test_frag_policy(dst, timeout=1):
    src = dnet.intf().get_dst(dst)['addr']
    
    sock = socket.socket(socket.AF_INET, socket.SOCK_RAW, socket.IPPROTO_ICMP)
    sock.setsockopt(socket.IPPROTO_IP, socket.IP_HDRINCL, 1)
    sock.settimeout(timeout)
    
    off_data = [
        [ dpkt.ip.IP_MF | 0, dpkt.icmp.ICMP(type=8,
            data=dpkt.icmp.ICMP.Echo(id=666, data='1'*3*8)) ],
        [ dpkt.ip.IP_MF | (1 + 4), '2'*2*8 ],
        [ dpkt.ip.IP_MF | (1 + 6), '3'*3*8 ],
        [ dpkt.ip.IP_MF | (1 + 1), '4'*4*8 ],
        [ dpkt.ip.IP_MF | (1 + 6), '5'*3*8 ],
        [ (1 + 9), '6'*3*8 ]
        ]
    policy_sum_data = (
        ( 'BSD', 19388, '1'*3*8 + '4'*2*8 + '2'*8 + '3'*3*8 + '6'*3*8 ),
        ( 'BSD-right', 9108, '1'*8 + '4'*3*8 + '2'*2*8 + '5'*3*8 + '6'*3*8 ),
        ( 'Linux', 13220, '1'*3*8 + '4'*2*8 + '2'*8 + '5'*3*8 + '6'*3*8 ),
        ( 'First', 21444, '1'*3*8 + '4'*8 + '2'*2*8 + '3'*3*8 + '6'*3*8 ),
        ( 'Last', 7052, '1'*8 + '4'*4*8 + '2'*8 + '5'*3*8 + '6'*3*8 )
        )
        
    ip = dpkt.ip.IP(src=src.ip, dst=dst.ip, p=dpkt.ip.IP_PROTO_ICMP)
    match = None
    for policy, sum, data in policy_sum_data:
        icmp = off_data[0][1]
        ip.id = icmp.sum = sum
        for ip.off, ip.data in off_data:
            ip.len = 20 + len(ip.data)
            ip.sum = 0
            sock.sendto(str(ip), (str(dst), 1))
        try:
            rip = dpkt.ip.IP(sock.recv(1500))
            if rip.icmp.type == 0 and rip.icmp.echo.id == 666:
                match = policy
                break
        except socket.timeout:
            pass
    if not match:
        print 'no ICMP echo response, exiting'
        raise SystemExit
    elif rip.icmp.echo.data == data:
        print 'IP frag: %s' % match
    else: # XXX - IOS!
        print 'IP frag: %s, with mismatch: %r' % (match, rip.icmp.echo.data)
        
class RawTcpConn(object):
    """Sleazy raw TCP connection handler."""
    def __init__(self, dst, p, dport):
        self.rand = dnet.rand()
        dstif = dnet.intf().get_dst(dst)
        name, src = dstif['name'], dstif['addr']
        sport = self.rand.uint16()
        
        self.fw = dnet.fw()
        self.rawip = dnet.ip()
        self.pcap = pcap.pcap(name, timeout_ms=1000, promisc=False)
        self.pcap.setfilter('src %s and proto %d and '
                            'src port %d and dst port %d' %
                            (dst, p, dport, sport))
        self.rule = { 'device':name,
                      'op':dnet.FW_OP_BLOCK, 'dir':dnet.FW_DIR_IN,
                      'src':dst, 'dst':src, 'proto':p,
                      'sport':(dport, dport), 'dport':(sport, sport) }
        self.fw.add(self.rule)
        self.pkts = []
        
        self.ip = dpkt.ip.IP(id=self.rand.uint16(),
                             src=src.ip, dst=dst.ip, p=p, len=40)
        self.tcp = dpkt.tcp.TCP(seq=self.rand.uint32(),
                                sport=sport, dport=dport)
        self.ip.data = self.tcp
        self.rawip.send(str(self.ip))
        
        self._recv()
        if self.rip.tcp.flags & dpkt.tcp.TH_SYN:
            self.tcp.flags = dpkt.tcp.TH_ACK
            self._send()
        else:
            self.close()
            raise IOError
    
    def _send(self, off=0):
        self.tcp.seq = self.rip.tcp.ack + off
        if self.rip.tcp.flags & (dpkt.tcp.TH_SYN|dpkt.tcp.TH_FIN) == 0:
            self.tcp.ack = self.rip.tcp.seq + len(self.rip.tcp.data)
        else:
            self.tcp.ack = self.rip.tcp.seq + 1
        self.ip.id = self.rand.uint16()
        self.ip.len = 20 + len(self.tcp)
        self.ip.sum = self.tcp.sum = 0
        self.rawip.send(str(self.ip))
    
    def _recv(self):
        #while not self.pkts:
        if not self.pkts:
            self.pkts = self.pcap.readpkts()
            if not self.pkts:
                raise IOError
        self.rip = dpkt.ip.IP(self.pkts.pop(0)[1][self.pcap.dloff:])
    
    def send(self, buf, off=0):
        self.tcp.flags = dpkt.tcp.TH_ACK
        self.tcp.data = buf
        self._send(off)

    def close(self):
        self.tcp.flags = dpkt.tcp.TH_RST | dpkt.tcp.TH_ACK
        self.tcp.data = ''
        self._send()
    
    def recv(self):
        while True:
            self._recv()
            if self.rip.tcp.data:
                return self.rip.tcp.data
    
    def __del__(self):
        self.fw.delete(self.rule)

def test_seg_policy(dst):
    try:
        sock = RawTcpConn(dst, dpkt.ip.IP_PROTO_TCP, 22)
    except IOError:
        print 'tcp/22 not open, exiting'
        raise SystemExit

    vers = sock.recv()
    sock.send(vers[1:4], 1)
    sock.send('X' + vers[4:], 3)
    sock.send(vers[0], 0)

    res = sock.recv()
    if res != 'Protocol mismatch.\n':
        sock.close()
        sock = RawTcpConn(dst, dpkt.ip.IP_PROTO_TCP, 22)
        vers = sock.recv()
        sock.send(vers[1:3], 1)
        sock.send('+', 8)
        sock.send('X' + vers[3:], 2)
        sock.send(vers[0], 0)
        
        res = sock.recv()
        if res == 'Protocol mismatch.\n':
            print 'TCP seg: First'
        else:
            print 'TCP seg: BSD'
    else:
        print 'TCP seg: Last'

    sock.close()

def main():
    try:
        dst = dnet.addr(sys.argv[1])
    except:
        print >>sys.stderr, 'usage: %s <dst>' % sys.argv[0]
        sys.exit(1)

    test_frag_policy(dst)
    test_seg_policy(dst)

if __name__ == '__main__':
    main()
